(*
  Ita IDE Plugin

  Copyright (c) 2014 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit UMain;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Generics.Collections,
  System.IniFiles, System.Win.Registry, Vcl.Controls, Vcl.Graphics, Vcl.Menus,
  Vcl.Imaging.jpeg, Vcl.Imaging.pngimage, ToolsAPI, UConfig;

procedure LoadBackgroundImage;

var
  ImageFileName: string = 'bgimage.png';
  ImagePos: TPoint = (X: 600; Y: 0);
  ImageAlpha: Byte;

implementation

const
  EVFillRectCodes: array[0..$29] of Byte = (
    $53, $56, $8B, $F2, $8B, $D8, $8B, $83, $FF, $FF, $FF, $FF, $8B, $40, $FF, $E8,
    $FF, $FF, $FF, $FF, $50, $56, $8B, $83, $FF, $FF, $FF, $FF, $E8, $FF, $FF, $FF,
    $FF, $50, $E8, $FF, $FF, $FF, $FF, $5E, $5B, $C3
  );
  EVScrollRectCodes: array[0..5] of Byte = (
    $55,          //PUSH EBP
    $8B, $EC,     //MOV EBP,ESP
    $83, $C4, $FF //ADD ESP,-28 (0xD8)
  );

type
  TEVFillRectCodes = packed record
    PUSH_EBX: Byte; //0x53
    PUSH_ESI: Byte; //0x56
    MOV_ESI_EDX: Word; //0x8B 0xF2
    MOV_EBX_EAX: Word; //0x8B 0xD8
    MOV_EAX_DWORD_OP: Word; //0x8B 0x83
    Canvas_Offset: Integer;
    MOV_EAX_DWORD_OP2: Word; //0x8B 0x40
    BrushOffset: Byte;
    CALL_TBrush_GetHandle_OP: Byte; //0xE8
    TBrush_GetHandle_Offset: Integer; //vcl160.Vcl::Graphics::TBrush::GetHandle
    PUSH_EAX: Byte; //0x50
    PUSH_ESI2: Byte; //0x56
    MOV_EAX_DWORD_OP3: Word; //0x8B 0x83
    Canvas_Offset2: Integer;
    CALL_TCanvas_GetHandle_OP: Byte; //0xE8
    TCanvas_GetHandle_Offset: Integer; //vcl160.Vcl::Graphics::TCanvas::GetHandle
    PUSH_EAX2: Byte; //0x50
    CALL_FillRect_OP: Byte; //0xE8
    FillRect_Offset: Integer; //user32.FillRect
    POP_ESI: Byte; //0x5E
    POP_EBX: Byte; //0x5B
    RETN: Byte; //0xC3
  end;
  PEVFillRectCodes = ^TEVFillRectCodes;

var
  sRTLVersion: string;
  EVFillRect_ContinueAddr: Pointer;
  EVScrollRect_ContinueAddr: Pointer;

  ImageSize: TSize;
  BufferDict: TObjectDictionary<TColor,TBitmap>;

  TCustomEditControl_EVScrollRect: Pointer;
  TCustomEditControl_EVFillRect: Pointer;

function GetModuleName(hModule: THandle): string;
var
  Buffer: array[0..260] of Char;
begin
  SetString(Result, Buffer, GetModuleFileName(hModule, Buffer, Length(Buffer)))
end;

function MyEVFillRect(Self: Pointer; const R: TRect): Boolean;
var
  codes: PEVFillRectCodes absolute TCustomEditControl_EVFillRect;
  canvas: TCanvas;
  rt: TRect;
  buf: TBitmap;
begin
  canvas := PPointer(Integer(Self) + codes^.Canvas_Offset)^;
  rt := Bounds(ImagePos.X, ImagePos.Y, ImageSize.cx, ImageSize.cy);

  Result := IntersectRect(rt, rt, R);
  if Result then
  try
    FillRect(canvas.Handle, R, canvas.Brush.Handle);
    if BufferDict.TryGetValue(canvas.Brush.Color, buf) then
    begin
      canvas.CopyRect(rt, buf.Canvas, Bounds(rt.Left-ImagePos.X, rt.Top-ImagePos.Y, rt.Right-rt.Left, rt.Bottom-rt.Top));
    end;
  except
  end;
end;

procedure HookEVFillRect;
asm
  PUSH EAX
  PUSH EDX
  CALL MyEVFillRect
  TEST AL,AL
  POP EDX
  POP EAX
  JNE @@1

  PUSH EBX
  PUSH ESI
  MOV ESI,EDX
  MOV EBX,EAX
  JMP EVFillRect_ContinueAddr
@@1:
end;

procedure HookEVScrollRect(Self: Pointer; const Src, Dest: TRect; DX, DY: Integer);
begin
  TWinControl(Self).Invalidate;
end;

procedure RefreshEditors;
var
  iModule, iFile, iView: Integer;
  service: IOTAModuleServices;
  module: IOTAModule;
  editor: IOTASourceEditor;
  window: INTAEditWindow;

  procedure EnumControls(Control: TWinControl; Level: Integer = 0);
  var
    iFile: Integer;
  begin
    if Control.ClassName = 'TEditControl' then
    begin
      Control.Invalidate;
      Exit;
    end;
    for iFile := 0 to Control.ComponentCount-1 do
    begin
      if Control.Components[iFile] is TWinControl then
        EnumControls(TWinControl(Control.Components[iFile]), Level+1);
    end;
  end;

begin
  try
    service := BorlandIDEServices as IOTAModuleServices;
    for iModule := 0 to service.ModuleCount-1 do
    begin
      module := service.Modules[iModule];
      for iFile := 0 to module.ModuleFileCount-1 do
      begin
        if not Supports(module.ModuleFileEditors[iFile], IOTASourceEditor, editor) then Continue;
        for iView := 0 to editor.EditViewCount-1 do
        begin
          window := editor.EditViews[iView].GetEditWindow;
          if not Assigned(window) then Continue;
          if window.Form <> nil then
          begin
            EnumControls(window.Form);
          end;
        end;
      end;
    end;
  except
  end;
end;

procedure Hook;
var
  oldProtect: Cardinal;

  procedure JmpHookProc(Target: Pointer; ReplaceProc: Pointer);
  begin
    VirtualProtect(Target, 5, PAGE_READWRITE, oldProtect);
    PByte(Target)^ := $E9;
    PInteger(Integer(Target)+1)^ := Integer(ReplaceProc) - Integer(Target) - 5;
    VirtualProtect(Target, 5, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

  procedure CallHookProc(Target: Pointer; ReplaceProc: Pointer);
  begin
    VirtualProtect(Target, 5, PAGE_READWRITE, oldProtect);
    PByte(Target)^ := $E8;
    PInteger(Integer(Target)+1)^ := Integer(ReplaceProc) - Integer(Target) - 5;
    VirtualProtect(Target, 5, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

begin
  EVFillRect_ContinueAddr := Pointer(Integer(TCustomEditControl_EVFillRect) + 6);
  JmpHookProc(TCustomEditControl_EVFillRect, @HookEVFillRect);

  EVScrollRect_ContinueAddr := Pointer(Integer(TCustomEditControl_EVScrollRect) + 6);
  JmpHookProc(TCustomEditControl_EVScrollRect, @HookEVScrollRect);

  RefreshEditors;
end;
                                                                                  
procedure Unhook;

  procedure UnhookProc(Target: Pointer; Size: Integer; const OrgCodes: array of Byte);
  var
    oldProtect: Cardinal;
  begin
    VirtualProtect(Target, Size, PAGE_READWRITE, oldProtect);
    Move(OrgCodes[0], Target^, Size);
    VirtualProtect(Target, Size, oldProtect, oldProtect);
    FlushInstructionCache(GetCurrentProcess, Target, 5);
  end;

begin
  if TCustomEditControl_EVFillRect = nil then Exit;
  if TCustomEditControl_EVScrollRect = nil then Exit;

  UnhookProc(TCustomEditControl_EVFillRect, 5, EVFillRectCodes);
  UnhookProc(TCustomEditControl_EVScrollRect, 5, EVScrollRectCodes);

  RefreshEditors;
end;

function EnumModulesProc(HInstance: THandle; Data: Pointer): Boolean;
const
  sEVFillRect = '@Editorcontrol@TCustomEditControl@EVFillRect$qqrrx18System@Types@TRect';
  sEVScrollRect = '@Editorcontrol@TCustomEditControl@EVScrollRect$qqrp18System@Types@TRectt1ii';
var
  fname: string;
  i: Integer;

  procedure ClearPointers;
  begin
    TCustomEditControl_EVFillRect := nil;
    TCustomEditControl_EVScrollRect := nil;
  end;

  function IsValidCodes(P: PByte; const Codes: array of Byte): Boolean;
  var
    i: Integer;
  begin
    for i := Low(Codes) to High(Codes) do
    begin
      if P^ and Codes[i] <> p^ then
      begin
        ClearPointers;
        Result := False;
        Exit;
      end;
      Inc(p);
    end;
    Result := True;
  end;

begin
  fname := LowerCase(ExtractFileName(GetModuleName(HInstance)));
  if (Pos('coreide', fname) = 1) and (ExtractFileExt(fname) = '.bpl') then
  begin
    Result := False;
    sRTLVersion := Copy(fname, 8, Length(fname)-11);
    if not TryStrToInt(sRTLVersion, i) then Exit;

    TCustomEditControl_EVFillRect := GetProcAddress(HInstance, sEVFillRect);
    TCustomEditControl_EVScrollRect := GetProcAddress(HInstance, sEVScrollRect);

    if (TCustomEditControl_EVFillRect = nil) or (TCustomEditControl_EVScrollRect = nil) then
    begin
      ClearPointers;
      Exit;
    end;

    if not IsValidCodes(TCustomEditControl_EVFillRect, EVFillRectCodes) then Exit;
    if not IsValidCodes(TCustomEditControl_EVScrollRect, EVScrollRectCodes) then Exit;

    Hook;
  end
  else begin
    Result := True;
  end;
end;

procedure LoadSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ExtractFilePath(GetModuleName(HInstance)) + 'itaide.ini');
  try
    ImageFileName := ini.ReadString('Image', 'FileName', ImageFileName);
    ImagePos.X := ini.ReadInteger('Image', 'Left', ImagePos.X);
    ImagePos.Y := ini.ReadInteger('Image', 'Top', ImagePos.Y);
    ImageAlpha := ini.ReadInteger('Image', 'Alpha', High(Byte));
  finally
    ini.Free;
  end;
end;

procedure SaveSettings;
var
  ini: TMemIniFile;
begin
  ini := TMemIniFile.Create(ExtractFilePath(GetModuleName(HInstance)) + 'itaide.ini');
  try
    ini.WriteString('Image', 'FileName', ImageFileName);
    ini.WriteInteger('Image', 'Left', ImagePos.X);
    ini.WriteInteger('Image', 'Top', ImagePos.Y);
    ini.WriteInteger('Image', 'Alpha', ImageAlpha);
    ini.UpdateFile;
  finally
    ini.Free;
  end;
end;

procedure LoadBackgroundImage;

  procedure LoadImage(const Path: string; Image: TPngImage);
  var
    ext: string;
    png: TPngImage;
    jpg: TJPEGImage;
    bmp: TBitmap;
    i, x, y: Integer;
    TransparencyChunk: TChunktRNS;
    alpha: Double;
  begin
    ext := LowerCase(ExtractFileExt(Path));
    if ext = '.png' then
    begin
      png := TPNGImage.Create;
      try
        png.LoadFromFile(path);
        if png.Header.ColorType = COLOR_RGBALPHA then
        begin
          Image.Assign(png);
          if ImageAlpha <> 255 then
          begin
            alpha := ImageAlpha / 255;
            for y := 0 to Image.Height-1 do
              for x := 0 to Image.Width-1 do
                Image.AlphaScanline[y]^[x] := Trunc(Image.AlphaScanline[y]^[x] * alpha);
          end;
        end
        else begin
          Image.Assign(png);
          case Image.Header.ColorType of
            COLOR_GRAYSCALE, COLOR_RGB:
            begin
              Image.CreateAlpha;
              for i := 0 to Image.Height-1 do
                FillChar(Image.AlphaScanline[i]^, Image.Width, ImageAlpha);
            end;
            COLOR_PALETTE:
            begin
              Image.CreateAlpha;
              TransparencyChunk := TChunktRNS(Image.Chunks.ItemFromClass(TChunktRNS));
              FillChar(TransparencyChunk.PaletteValues[0], 256, ImageAlpha);
            end;
          end;
        end;
      finally
        png.Free;
      end;
    end
    else if (ext = '.jpg') or (ext = '.jpeg') then
    begin
      jpg := TJPEGImage.Create;
      try
        jpg.LoadFromFile(path);
        png := TPngImage.CreateBlank(COLOR_RGBALPHA, 8, jpg.Width, jpg.Height);
        try
          png.Canvas.Draw(0, 0, jpg);
          for i := 0 to png.Height-1 do
            FillChar(png.AlphaScanline[i]^, png.Width, ImageAlpha);
          Image.Assign(png);
        finally
          png.Free;
        end;
      finally
        jpg.Free;
      end;
    end
    else if ext = '.bmp' then
    begin
      bmp := TBitmap.Create;
      try
        bmp.LoadFromFile(path);

        png := TPngImage.CreateBlank(COLOR_RGBALPHA, 8, bmp.Width, bmp.Height);
        try
          png.Canvas.Draw(0, 0, bmp);
          for i := 0 to png.Height-1 do
            FillChar(png.AlphaScanline[i]^, png.Width, ImageAlpha);
          Image.Assign(png);
        finally
          png.Free;
        end;
      finally
        bmp.Free;
      end;
    end
    else
      raise Exception.Create('Unsupported image format');
  end;

const
  RegKey =
    {$IFDEF VER280 XE7 }'\Software\Embarcadero\BDS\15.0\Editor\Highlight\'{$ENDIF}
    {$IFDEF VER270 XE6 }'\Software\Embarcadero\BDS\14.0\Editor\Highlight\'{$ENDIF}
    {$IFDEF VER260 XE5 }'\Software\Embarcadero\BDS\12.0\Editor\Highlight\'{$ENDIF}
    {$IFDEF VER250 XE4 }'\Software\Embarcadero\BDS\11.0\Editor\Highlight\'{$ENDIF}
    {$IFDEF VER240 XE3 }'\Software\Embarcadero\BDS\10.0\Editor\Highlight\'{$ENDIF}
    {$IFDEF VER230 XE2 }'\Software\Embarcadero\BDS\9.0\Editor\Highlight\'{$ENDIF}
  ;
  RegName = 'Background Color New';
var
  Image: TPngImage;
  path: string;
  reg: TRegistry;
  buf: TBitmap;
  keys: TStringList;
  i: Integer;
  s: string;
  color: TColor;
begin
  BufferDict.Clear;
  
  path := ExtractFilePath(GetModuleName(HInstance)) + ImageFileName;
  if not FileExists(path) then
  begin
    path := ImageFileName;
    if not FileExists(path) then Exit;
  end;

  try
    Image := TPngImage.Create;
    try
      LoadImage(path, Image);

      ImageSize.cx := Image.Width;
      ImageSize.cy := Image.Height;

      reg := TRegistry.Create;
      try
        reg.RootKey := HKEY_CURRENT_USER;
        if not reg.OpenKeyReadOnly(RegKey) then Exit;

        keys := TStringList.Create;
        try
          reg.GetKeyNames(keys);
          for i := 0 to keys.Count-1 do
          begin
            if not reg.OpenKeyReadOnly(RegKey + keys[i]) then Continue;

            s := reg.ReadString(RegName);
            color := StringToColor(s);

            if BufferDict.ContainsKey(color) then Continue;

            buf := TBitmap.Create;
            try
              buf.Width := Image.Width;
              buf.Height := Image.Height;
              buf.Canvas.Brush.Color := color;
              buf.Canvas.FillRect(Rect(0, 0, buf.Width, buf.Height));
              buf.Canvas.Draw(0, 0, Image);
              BufferDict.Add(color, buf);
            except
              buf.Free;
            end;
          end;
        finally
          keys.Free;
        end;
      finally
        reg.Free;
      end;
    finally
      Image.Free;
    end;
  except
    BufferDict.Clear;
    Exit;
  end;

  RefreshEditors;
end;

var
  ItaIDEMenu: TMenuItem;

procedure ItaIDEImageMenuClick(Self: TMenuItem; Sender: TMenuItem);
begin
  ShowConfigDialog;
end;

procedure RegisterMenu;
var
  nta: INTAServices;
  i, j: Integer;
  method: TMethod;
begin
  nta := BorlandIDEServices as INTAServices;
  for i := 0 to nta.MainMenu.Items.Count-1 do
  begin
    if nta.MainMenu.Items[i].Name <> 'ToolsMenu' then Continue;

    for j := 0 to nta.MainMenu.Items[i].Count-1 do
    begin
      if nta.MainMenu.Items[i][j].Name <> 'ToolsToolsItem' then Continue;

      ItaIDEMenu := TMenuItem.Create(nil);
      ItaIDEMenu.AutoHotkeys := maManual;
      ItaIDEMenu.Caption := 'IDEݒ';
      method.Code := @ItaIDEImageMenuClick;
      method.Data := ItaIDEMenu;
      ItaIDEMenu.OnClick := TNotifyEvent(method);
      nta.MainMenu.Items[i].Insert(j, ItaIDEMenu);
      Break;
    end;
    Break;
  end;
end;

procedure UnregisterMenu;
begin
  if ItaIDEMenu <> nil then
  begin
    ItaIDEMenu.Parent.Remove(ItaIDEMenu);
    ItaIDEMenu.Free;
  end;
end;

initialization
  BufferDict := TObjectDictionary<TColor,TBitmap>.Create([doOwnsValues]);
  LoadSettings;
  LoadBackgroundImage;
  EnumModules(EnumModulesProc, nil);
  RegisterMenu;
finalization
  UnregisterMenu;
  Unhook;
  BufferDict.Free;
end.
